from .base import BaseUnitProxy, BaseWorkManager
from .noop import NoOpUnitProxy, NoOpWorkManager

__all__ = (
    "BaseUnitProxy",
    "BaseWorkManager",
    "NoOpUnitProxy",
    "NoOpWorkManager",
)



from ..result import OUT, Result
from ..units import CONTEXT, BaseUnit
from .base import BaseUnitProxy, BaseWorkManager


class NoOpUnitProxy(BaseUnitProxy[CONTEXT, OUT]):  # pragma: no cover

    def __init__(
        self,
        *,
        unit: BaseUnit[CONTEXT, OUT],
    ) -> None:
        self._unit = unit

    async def __call__(self, context: CONTEXT) -> Result[OUT]:
        return await self._unit(context)


class NoOpWorkManager(BaseWorkManager):  # pragma: no cover
    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> BaseUnitProxy[CONTEXT, OUT]:
        return NoOpUnitProxy(unit=unit)



import typing as t
from dataclasses import dataclass
from datetime import datetime, timezone

from ..types import MISSING

ENTITY_ID = t.TypeVar("ENTITY_ID", bound=t.Any)
ENTITY_TYPE = t.TypeVar("ENTITY_TYPE", bound="Entity[t.Any]")


@dataclass(frozen=True)
class Entity(t.Generic[ENTITY_ID]):
    id: ENTITY_ID


@dataclass(frozen=True)
class AuditedEntity(Entity[ENTITY_ID]):
    created_date: datetime = MISSING  # type: ignore[assignment]
    updated_date: datetime = MISSING  # type: ignore[assignment]

    def __post_init__(self) -> None:
        now = datetime.now(tz=timezone.utc)

        if self.created_date is MISSING:  # pragma: no cover
            object.__setattr__(self, "created_date", now)
        if self.updated_date is MISSING:  # pragma: no cover
            object.__setattr__(self, "updated_date", now)



from ...result import OUT, Result
from ...units import CONTEXT, BaseUnit
from ...work import BaseWorkManager
from ...work.transactional import BaseTransactionManager
from ...work.transactional.base import TRANSACTION
from ..base import BaseUnitProxy


class TransactionalUnitProxy(BaseUnitProxy[CONTEXT, OUT]):
    def __init__(
        self,
        *,
        transaction_manager: BaseTransactionManager[TRANSACTION],
        unit: BaseUnit[CONTEXT, OUT],
    ) -> None:
        self._transaction_manager = transaction_manager
        self._unit = unit

    async def __call__(self, context: CONTEXT) -> Result[OUT]:
        async with self._transaction_manager.transaction() as trx:
            result = await self._unit(context)

            if result.is_error():
                await trx.rollback()
            else:
                await trx.commit()

            return result


class TransactionalWorkManager(BaseWorkManager):
    def __init__(
        self,
        *,
        transaction_manager: BaseTransactionManager[TRANSACTION],
    ) -> None:
        self._transaction_manager = transaction_manager

    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> BaseUnitProxy[CONTEXT, OUT]:
        return TransactionalUnitProxy(
            transaction_manager=self._transaction_manager,
            unit=unit,
        )



import abc
from abc import ABC

from ..result import OUT, Result
from ..units import CONTEXT, BaseUnit


class BaseUnitProxy(BaseUnit[CONTEXT, OUT], ABC):
    async def do_with(self, context: CONTEXT) -> Result[OUT]:
        return await self(context)


class BaseWorkManager(ABC):
    @abc.abstractmethod
    def by(self, unit: BaseUnit[CONTEXT, OUT]) -> BaseUnitProxy[CONTEXT, OUT]:
        raise NotImplementedError
